/* $Id: mod_accounting.c,v 1.5 2002/01/05 11:50:50 tellini Exp $ */

/*
	Copyright (c) 2001-2002 Simone Tellini

	Permission is hereby granted, free of charge, to any person obtaining 
	a copy of this software and associated documentation files (the "Software"), 
	to deal in the Software without restriction, including without limitation 
	the rights to use, copy, modify, merge, publish, distribute, sublicense, 
	and/or sell copies of the Software, and to permit persons to whom the 
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included 
	in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
	THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
	DEALINGS IN THE SOFTWARE.
*/

//#define DEBUG

#include <stdio.h>
#include <stdlib.h>

#include "mod_accounting.h"

module MODULE_VAR_EXPORT accounting_module;

typedef int ( *DBQuery )( accounting_state *cfg, server_rec *server, pool *p, char *query );
typedef int ( *DBSetup )( accounting_state *cfg );
typedef void ( *DBClose )( accounting_state *cfg );

typedef struct _DBHandler
{
	char		*ID;
	DBSetup		Setup;
	DBClose		Close;
	DBQuery		Query;
} DBHandler;

static const DBHandler DBDrivers[] =
{
#ifdef NEED_POSTGRES
	{ "postgres", PgSetup, PgClose, PgQuery },
#endif
#ifdef NEED_MYSQL
	{ "mysql", MySetup, MyClose, MyQuery },
#endif
};

#define DB_MAX ( sizeof( DBDrivers ) / sizeof( DBDrivers[0] ))

// ------------------- UTILITY FUNCS -----------------

#ifdef DEBUG
request_rec *rr;
#endif

// hook function pass to ap_table_do()
static int GetHeaderLen( long *count, const char *key, const char *val )
{
	*count += strlen( key ) + strlen( val ) + 4; // 4 for ": " + CR + LF

#ifdef DEBUG
	ap_log_error( APLOG_MARK, ERRLEVEL, rr->server, 
				  ap_pstrcat( rr->pool, key, ": ", val, NULL ));
#endif

	return( 1 );
}

// computes the length of a table
static long TableLen( request_rec *r, table *tab )
{
	long count = 0;

	ap_table_do((int (*) (void *, const char *, const char *)) GetHeaderLen, (void *) &count, tab, NULL );

	return( count );
}

// computes the number of bytes sent
static long BytesSent( request_rec *r )
{
	long sent;

#ifdef DEBUG
	ap_log_error( APLOG_MARK, ERRLEVEL, rr->server, 
				  "BytesSent" );
#endif

	sent = TableLen( r, r->headers_out ) + TableLen( r, r->err_headers_out ) + 2 +	// 2 for CRLF
		   11 + strlen( r->status_line ) +											// HTTP/1.x nnn blah
		   10 + strlen( ap_get_server_version() ) +									// Server: line
		   8 + strlen( ap_gm_timestr_822( r->pool, r->request_time ));				// Date: line

    if(( sent >= 255 ) && ( sent <= 257 ))
        sent += sizeof( "X-Pad: avoid browser bug" ) + 1;

	if( r->sent_bodyct ) {
		long int bs;

		// this is more accurate than bytes_sent in presence of modules
		// which manipulate the output (eg. mod_gzip)
		ap_bgetopt( r->connection->client, BO_BYTECT, &bs );
		
		sent += bs;
	}

	return( sent );
}

// computes the number of bytes received
static long BytesRecvd( request_rec *r )
{
	long		recvd;
	const char *len;

#ifdef DEBUG
	ap_log_error( APLOG_MARK, ERRLEVEL, rr->server, 
				  "BytesRecvd" );
#endif

	recvd = strlen( r->the_request ) + TableLen( r, r->headers_in ) + 4; // 2 for CRLF after the request, 2 for CRLF after all headers

	len = ap_table_get( r->headers_in, "Content-Length" );

	if( len )
		recvd += atol( len );

	return( recvd );
}

// ------------------- CONFIG -----------------

static char *set_db( cmd_parms *parms, void *dummy, char *arg )
{
	accounting_state *cfg = (accounting_state *) ap_get_module_config( parms->server->module_config, &accounting_module );

	cfg->DBName = arg;

	return( NULL );
}

static char *set_driver( cmd_parms *parms, void *dummy, char *arg )
{
	accounting_state *cfg = (accounting_state *) ap_get_module_config( parms->server->module_config, &accounting_module );
	int				 i, found = 0;
	char			 *ptr = arg;

	while( *ptr )
		*ptr++ = tolower( *ptr );

	// let's see if we have the requested driver
	for( i = 0; i < DB_MAX; i++ )
		if( !strcmp( DBDrivers[ i ].ID, arg )) {
			cfg->DBDriver = i;
			found         = 1;
			break;
		}

	return( found ? NULL : "wrong DB driver" );
}

static char *set_db_host( cmd_parms *parms, void *dummy, char *host, char *port )
{
	accounting_state *cfg = (accounting_state *) ap_get_module_config( parms->server->module_config, &accounting_module );

	cfg->DBHost = host;
	cfg->DBPort = port;

	return( NULL );
}

static char *set_login_info( cmd_parms *parms, void *dummy, char *user, char *pwd )
{
	accounting_state *cfg = (accounting_state *) ap_get_module_config( parms->server->module_config, &accounting_module );

	cfg->DBUser = user;
	cfg->DBPwd  = pwd;

	return( NULL );
}

static char *set_query_fmt( cmd_parms *parms, void *dummy, char *arg )
{
	accounting_state *cfg = (accounting_state *) ap_get_module_config( parms->server->module_config, &accounting_module );

	cfg->QueryFmt = arg;

	return( NULL );
}

static char *set_timed_updates( cmd_parms *parms, void *dummy, char *arg )
{
    accounting_state *cfg = (accounting_state *) ap_get_module_config( parms->server->module_config, &accounting_module );

    cfg->UpdateTimeSpan = atoi( arg );

    return( NULL );
}

/* Setup of the available httpd.conf configuration commands.
 * command, function called, NULL, where available, how many arguments, verbose description
 */
typedef const char *(*hook_func)();

static command_rec acct_cmds[] = 
{
	{ "AccountingQueryFmt", (hook_func) set_query_fmt, NULL, RSRC_CONF, TAKE1,
	  "The query to execute to log the transactions. Available placeholders are %s for sent bytes, %r for received ones, %h for the virtual host name." },

	{ "AccountingDatabase", (hook_func) set_db, NULL, RSRC_CONF, TAKE1,
	  "The name of the database for logging" },

	{ "AccountingDatabaseDriver", (hook_func) set_driver, NULL, RSRC_CONF, TAKE1,
	  "The kind of the database for logging" },

	{ "AccountingDBHost", (hook_func) set_db_host, NULL, RSRC_CONF, TAKE2,
	  "Host and port needed to connect to the database" },

	{ "AccountingLoginInfo", (hook_func) set_login_info, NULL, RSRC_CONF, TAKE2,
	  "User and password required for logging into the database" },

	{ "AccountingTimedUpdates", (hook_func) set_timed_updates, NULL, RSRC_CONF, TAKE1,
	  "Number of seconds to wait between 2 update queries (performance tuning)" },

	{ NULL }
};


// ------------------- HOOKS -----------------
	
/* Set up space for the various major configuration options */
static void *acct_make_state( pool *p, server_rec *s )
{
	accounting_state *cfg = ( accounting_state * ) ap_palloc( p, sizeof( accounting_state ));

	memset( cfg, 0, sizeof( *cfg ));

	return( cfg );
}

/* Routine to perform the actual construction and execution of the relevant
 * query
 */
static void *do_query( accounting_state *cfg, pool *p, server_rec *server )
{
	if(!( *DBDrivers[ cfg->DBDriver ].Setup )( cfg )) {

		ap_log_error( APLOG_MARK, ERRLEVEL, server, "Accounting: couldn't setup the database link!" );

	} else {
		char *query = "", *ptr = cfg->QueryFmt;

		// build the query string from the template
		while( ptr ) {
			char *next;
						
			next = strchr( ptr, '%' );

			if( next ) {
				char	tmp[ 2 ], sent[ 32 ], recvd[ 32 ];
				
				sprintf( sent, "%ld", cfg->Sent );
				sprintf( recvd, "%ld", cfg->Received );

				*next++ = '\0';

				switch( *next++ ) {

					case 'h':
						query = ap_pstrcat( p, query, ptr, server->server_hostname, NULL );
						break;

					case 's':
						query = ap_pstrcat( p, query, ptr, sent, NULL );
						break;

					case 'r':
						query = ap_pstrcat( p, query, ptr, recvd, NULL );
						break;

					default:
						tmp[0] = next[ -1 ];
						tmp[1] = '\0';

						query = ap_pstrcat( p, query, ptr, tmp, NULL );
						break;
				}

				next[ -2 ] = '%';

			} else
				query = ap_pstrcat( p, query, ptr, NULL );

			ptr = next;
		}

		( *DBDrivers[ cfg->DBDriver ].Query )( cfg, server, p, query );

		cfg->Received = cfg->Sent = 0;
	}
}

static int acct_transaction( request_rec *orig )
{
	accounting_state	*cfg = ap_get_module_config( orig->server->module_config, &accounting_module );
	request_rec			*r = orig;
	int					do_it = 1;

	// get to the last of the chain, we need the correct byte counters ;-)
	while( r->next )
		r = r->next;

#ifdef DEBUG
	rr = r;
#endif

	cfg->Received += BytesRecvd( r );
	cfg->Sent	  += BytesSent( r );

	if( cfg->UpdateTimeSpan ) { // group updates?
		time_t	now;

		time( &now );

		if( now - cfg->LastUpdate < cfg->UpdateTimeSpan )
			do_it = 0;
		else
			cfg->LastUpdate = now;
	}

	if( do_it )
		do_query( cfg, r->pool, r->server );

	return( OK );
}

/* Called on the exit of an httpd child process */
static void acct_child_exit( server_rec *s, pool *p )
{
	accounting_state *cfg = ap_get_module_config( s->module_config, &accounting_module );

	if( cfg->Sent || cfg->Received )
		do_query( cfg, p, s );

	( *DBDrivers[ cfg->DBDriver ].Close )( cfg );
}

// ------------------- MOD CONFIG -----------------

/* The configuration array that sets up the hooks into the module. */
module accounting_module = 
{
	STANDARD_MODULE_STUFF,
	NULL,					 /* initializer */
	NULL,					 /* create per-dir config */
	NULL,					 /* merge per-dir config */
	acct_make_state,		 /* server config */
	NULL,					 /* merge server config */
	acct_cmds,				 /* command table */
	NULL,					 /* handlers */
	NULL,					 /* filename translation */
	NULL,					 /* check_user_id */
	NULL,					 /* check auth */
	NULL,					 /* check access */
	NULL,					 /* type_checker */
	NULL,					 /* fixups */
	acct_transaction,		 /* logger */
#if MODULE_MAGIC_NUMBER >= 19970103
	NULL,					 /* header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
	NULL,                    /* child_init */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
	acct_child_exit,		/* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
	NULL					 /* [#0] post read-request */
#endif  
};
